import click
import uuid
from . import cmd_utils as utl


@click.group()
@click.pass_context
@click.option("--json_output", is_flag=True, default=False)
def instance(ctx, json_output):
    ctx.obj["JSON_OUTPUT"] = json_output


@click.command()
@click.pass_context
def types(ctx):
    utl.cli_message(ctx, "List instance types")
    instance_types = utl.do_get("instance-types", ctx)
    utl.cli_json_output(instance_types)


@click.command()
@click.pass_context
@click.option(
    "--state_file", default=None, help="Name of the file to persist instances info"
)
def show_all(ctx, state_file):
    utl.cli_message(ctx, "List running instances")
    all_instances = utl.do_get("instances", ctx)
    utl.cli_json_output(all_instances)
    if state_file:
        utl.save_instance_state(all_instances, state_file)


@click.command()
@click.pass_context
@click.option("--id", required=True, help="Id of the instance to check the status for")
def show_details(ctx, id):
    utl.cli_message(ctx, f"Show details for instance {id}")
    instance_details = utl.get_instance_details(ctx, id)
    utl.cli_json_output(instance_details)


@click.command()
@click.pass_context
@click.option("--id", default=None, help="Id of the instance to check the status for")
@click.option("--status", default="active", help="Name of the status to wait for")
def check_status(ctx, id, status):
    if id:
        utl.check_instance_status(ctx, id, status)
    else:
        all_instances = utl.do_get("instances", ctx)
        if not all_instances["data"]:
            utl.cli_message(ctx, "No instances available", force=True)
            return
        ids = list(map(lambda info: info["id"], all_instances["data"]))
        for id in ids:
            utl.check_instance_status(ctx, id, status)


@click.command()
@click.option("--name", default=None, help="Name of the new instance")
@click.option("--type", default="gpu_1x_a10", help="Instance Type")
@click.option("--region", default="us-east-1", help="Region")
@click.option("--ssh_key", required=True, help="SSH Key to access the instance")
@click.option("--qty", default=1, help="Number of instances")
@click.option(
    "--fs_name", default=None, help="A file system name to be attached to the instance"
)
@click.pass_context
def launch(ctx, name, type, region, ssh_key, qty, fs_name):
    payload = {
        "region_name": region,
        "instance_type_name": type,
        "ssh_key_names": [ssh_key],
        "quantity": qty,
        "name": name if name else str(uuid.uuid4()),
    }
    if fs_name:
        payload["file_system_names"] = [fs_name]

    utl.cli_message(ctx, f"Launch instance with parameters {payload}")

    launched_instance = utl.do_post("instance-operations/launch", ctx, payload)
    utl.cli_json_output(launched_instance)


@click.command()
@click.option(
    "--instance_id",
    required=True,
    help="Instance id to terminate",
)
@click.pass_context
# TODO change to support a list of ids
def terminate(ctx, instance_id):
    utl.cli_message(ctx, f"Terminate instance {instance_id}")
    payload = {"instance_ids": [instance_id]}
    terminated_instance = utl.do_post("instance-operations/terminate", ctx, payload)
    utl.cli_json_output(terminated_instance)


@click.command()
@click.option(
    "--instance_id",
    required=True,
    help="Instance id to terminate",
)
@click.pass_context
def restart(ctx, instance_id):
    utl.cli_message(f"Restarting instance {instance_id}")
    payload = {"instance_ids": [instance_id]}
    response = utl.do_post("instance-operations/restart", ctx, payload)
    restarted_instance = utl.get_json_or_raise(
        response, "Unable to restart instance {instance_id}"
    )
    utl.cli_json_output(restarted_instance)


instance.add_command(types)
instance.add_command(show_all)
instance.add_command(show_details)
instance.add_command(check_status)
instance.add_command(launch)
instance.add_command(terminate)
instance.add_command(restart)



import click
import requests
import json
import time

LAMBDA_LABS_ENDPOINT = "https://cloud.lambdalabs.com/api/v1"


class APIException(Exception):
    def __init__(self, http_status, body, message, details={}):
        super().__init__(message)
        self.http_status = http_status
        self.body = body
        self.message = message
        self.details = details

    def __str__(self):
        return f"API Exception: {self.message}\nstatus:{self.http_status}\nbody={self.body}\ndetails={self.details}"

    def __repr__(self):
        return f"APIException(http_status={self.http_status}, body={self.body}, details={self.details})"


def do_get(api_name, ctx):
    API_KEY = ctx.obj["API_KEY"]
    url = f"{LAMBDA_LABS_ENDPOINT}/{api_name}"
    headers = {"Authorization": f"Basic {API_KEY}"}
    response = requests.get(url, headers=headers)
    return get_json_or_raise(response, f"Unexpected exception calling {url}")


def do_post(api_name, ctx, payload, is_json=True):
    API_KEY = ctx.obj["API_KEY"]
    url = f"{LAMBDA_LABS_ENDPOINT}/{api_name}"
    headers = {"Authorization": f"Basic {API_KEY}"}
    if is_json:
        response = requests.post(url, headers=headers, json=payload)
    else:
        response = requests.post(url, headers=headers, data=payload)
    return get_json_or_raise(
        response, f"Unexpected exception calling {url}", 200, payload
    )


def do_delete(api_name, ctx):
    API_KEY = ctx.obj["API_KEY"]
    url = f"{LAMBDA_LABS_ENDPOINT}/{api_name}"
    headers = {"Authorization": f"Basic {API_KEY}"}
    response = requests.delete(url, headers=headers)
    return get_json_or_raise(response, f"Unexpected exception calling {url}", 200)


def cli_message(ctx, msg, force=False):
    json_out = ctx.obj["JSON_OUTPUT"]
    if force or not json_out:
        click.echo(msg)


def cli_json_output(data):
    if data:
        print(json.dumps(data))


def enable_http_debug():
    import logging
    import http.client as http_client

    http_client.HTTPConnection.debuglevel = 1

    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


def save_instance_state(instances_info, state_file, fields=["id", "name", "ip"]):
    infra_state = {"instances": []}
    for info in instances_info["data"]:
        infra_state["instances"].append(info)
    with open(state_file, "w") as state_file:
        state_file.write(json.dumps(infra_state))


def get_instance_details(ctx, id):
    response_json = do_get(f"instances/{id}", ctx)
    return response_json["data"]


def check_instance_status(ctx, id, status):
    max_retries, retry_gap_secs, retry_count = 6, 60, 0
    actual_status = "undefined"
    while retry_count < max_retries:
        instance_details = get_instance_details(ctx, id)
        cli_message(ctx, f"Instance {id} in {instance_details['status']} status")
        if instance_details["status"] == status:
            actual_status = status
            break
        retry_count += 1
        time.sleep(retry_gap_secs)
    cli_message(ctx, f"Instance {id} reached {actual_status} status")


def get_json_or_raise(response, message, expected_http_status=200, details={}):
    if response.status_code != expected_http_status:
        raise APIException(response.status_code, response.text, message, details)
    return response.json()


def get_text_or_raise(response, message, expected_http_status=200, details={}):
    if response.status_code != expected_http_status:
        raise APIException(response.status_code, response.text, message, details)
    return response.text



from lamblbs.instance_cmd import instance
from click.testing import CliRunner
from unittest.mock import patch
import lamblbs.cmd_utils as utl
import json
import pytest

API_KEY = "abc123"


@patch("lamblbs.cmd_utils.requests.get")
def test_show_all_instance_success(mock_get):
    mock_get.return_value.status_code = 200
    mock_get.return_value.json.return_value = {
        "data": [
            {
                "id": "1",
                "name": "training-node-1",
                "ip": "10.10.10.1",
                "status": "active",
                "ssh_key_names": ["macbook-pro"],
                "file_system_names": ["shared-fs"],
                "region": {"name": "us-tx-1", "description": "Austin, Texas"},
                "instance_type": {
                    "name": "gpu_1x_a100",
                    "description": "1x RTX A100 (24 GB)",
                    "price_cents_per_hour": 110,
                    "specs": {"vcpus": 24, "memory_gib": 800, "storage_gib": 512},
                },
                "hostname": "10-0-8-196.cloud.lambdalabs.com",
                "jupyter_token": "53968f128c4a4489b688c2c0a181d083",
                "jupyter_url": "https://...",
            }
        ]
    }

    runner = CliRunner()
    result = runner.invoke(
        instance, ["--json_output", "show-all"], obj={"API_KEY": API_KEY}
    )
    response = json.loads(result.output)
    assert len(response["data"]) == 1
    assert response["data"][0]["id"] == "1"
    assert response["data"][0]["name"] == "training-node-1"


@patch("lamblbs.cmd_utils.requests.get")
def test_types_instance_success(mock_get):
    mock_get.return_value.status_code = 200
    mock_get.return_value.json.return_value = {
        "data": {
            "gpu_1x_a100": {
                "instance_type": {
                    "name": "gpu_1x_a100",
                    "description": "1x RTX A100 (24 GB)",
                    "price_cents_per_hour": "80",
                    "specs": {"vcpus": 24, "memory_gib": 800, "storage_gib": 512},
                },
                "regions_with_capacity_available": [
                    {"name": "us-tx-1", "description": "Austin, Texas"}
                ],
            }
        }
    }

    runner = CliRunner()
    result = runner.invoke(
        instance, ["--json_output", "types"], obj={"API_KEY": API_KEY}
    )
    response = json.loads(result.output)
    assert response["data"]["gpu_1x_a100"]["instance_type"]["name"] == "gpu_1x_a100"


@patch("lamblbs.cmd_utils.requests.get")
def test_show_details_instance_success(mock_get):
    mock_get.return_value.status_code = 200
    mock_get.return_value.json.return_value = {
        "data": {
            "id": "1",
            "name": "training-node-1",
            "ip": "10.10.10.1",
            "status": "active",
            "ssh_key_names": ["macbook-pro"],
            "file_system_names": ["shared-fs"],
            "region": {"name": "us-tx-1", "description": "Austin, Texas"},
            "instance_type": {
                "name": "gpu_1x_a100",
                "description": "1x RTX A100 (24 GB)",
                "price_cents_per_hour": 110,
                "specs": {"vcpus": 24, "memory_gib": 800, "storage_gib": 512},
            },
            "hostname": "10-0-8-196.cloud.lambdalabs.com",
            "jupyter_token": "53968f128c4a4489b688c2c0a181d083",
            "jupyter_url": "https://jupyter...",
        }
    }

    runner = CliRunner()
    result = runner.invoke(
        instance,
        ["--json_output", "show-details", "--id", "1"],
        obj={"API_KEY": API_KEY},
    )
    response = json.loads(result.output)
    assert response["id"] == "1"


@patch("lamblbs.cmd_utils.requests.post")
def test_launch_instance_success(mock_post):
    mock_post.return_value.status_code = 200
    mock_post.return_value.json.return_value = {
        "data": [
            {
                "id": "1",
                "name": "training-node-1",
                "ip": "10.10.10.1",
                "status": "active",
                "ssh_key_names": ["macbook-pro"],
                "file_system_names": ["shared-fs"],
                "region": {"name": "us-tx-1", "description": "Austin, Texas"},
                "instance_type": {
                    "name": "gpu_1x_a100",
                    "description": "1x RTX A100 (24 GB)",
                    "price_cents_per_hour": 110,
                    "specs": {"vcpus": 24, "memory_gib": 800, "storage_gib": 512},
                },
                "hostname": "10-0-8-196.cloud.lambdalabs.com",
                "jupyter_token": "53968f128c4a4489b688c2c0a181d083",
                "jupyter_url": "https://jupyter...",
            }
        ]
    }

    runner = CliRunner()
    result = runner.invoke(
        instance,
        ["--json_output", "launch", "--ssh_key", "sample-ssh-key"],
        obj={"API_KEY": API_KEY},
    )
    response = json.loads(result.output)
    assert response["data"][0]["id"] == "1"


# TODO implement this unit test
@patch("lamblbs.cmd_utils.requests.get")
def test_check_status_details_instance_success(mock_get):
    pass


# TODO implement this unit test
@patch("lamblbs.cmd_utils.requests.get")
def test_restart_instance_success(mock_get):
    pass


# TODO implement this unit test
@patch("lamblbs.cmd_utils.requests.get")
def test_terminate_instance_success(mock_get):
    pass



from lamblbs.ssh_cmd import ssh
from click.testing import CliRunner
from unittest.mock import patch
import lamblbs.cmd_utils as utl
import json

API_KEY = "abc123"


@patch("lamblbs.cmd_utils.requests.get")
def test_list_ssh_success(mock_get):
    mock_get.return_value.status_code = 200
    mock_get.return_value.json.return_value = {
        "data": [
            {"id": "1", "name": "ssh-key-1", "public_key": "ssh-rsa AAAA001"},
            {"id": "2", "name": "ssh-key-2", "public_key": "ssh-rsa AAAAB02"},
        ]
    }

    runner = CliRunner()
    result = runner.invoke(ssh, ["--json_output", "list"], obj={"API_KEY": API_KEY})

    response = json.loads(result.output)
    assert len(response["data"]) == 2


@patch("lamblbs.cmd_utils.requests.get")
def test_list_ssh_fails(mock_get):
    mock_get.return_value.status_code = 401

    runner = CliRunner()
    result = runner.invoke(ssh, ["--json_output", "list"], obj={"API_KEY": API_KEY})
    assert 1 == result.exit_code
    assert isinstance(result.exception, utl.APIException)
    assert result.exception.http_status == 401


@patch("lamblbs.cmd_utils.requests.post")
def test_add_ssh_success(mock_post):
    mock_post.return_value.status_code = 200
    mock_post.return_value.json.return_value = {
        "data": {
            "id": "1",
            "name": "my-ssh-key",
            "public_key": "ssh-rsa AAAAB001",
            "private_key": "-----BEGIN RSA PRIVATE KEY-----\nKEY CONTENT-----END RSA PRIVATE KEY-----\n",
        }
    }

    runner = CliRunner()
    result = runner.invoke(
        ssh, ["--json_output", "add", "--name", "my-ssh-key"], obj={"API_KEY": API_KEY}
    )

    response = json.loads(result.output)
    assert response["data"]["id"] == "1"
    assert response["data"]["name"] == "my-ssh-key"


@patch("lamblbs.cmd_utils.requests.post")
def test_add_ssh_missing_name(mock_post):
    mock_post.return_value.status_code = 200

    runner = CliRunner()
    result = runner.invoke(ssh, ["--json_output", "add"], obj={"API_KEY": API_KEY})
    assert 2 == result.exit_code
    assert "Missing option '--name'" in result.output


@patch("lamblbs.cmd_utils.requests.delete")
def test_delete_ssh_success(mock_delete):
    mock_delete.return_value.status_code = 200
    mock_delete.return_value.json.return_value = {"data": {}}

    runner = CliRunner()
    result = runner.invoke(
        ssh, ["--json_output", "delete", "--id", "1"], obj={"API_KEY": API_KEY}
    )

    response = json.loads(result.output)
    assert response["data"] is not None



import click
import os

from . import instance_cmd
from . import ssh_cmd
from . import fs_cmd


@click.group()
@click.pass_context
def cli(ctx):
    ctx.obj["API_KEY"] = os.environ["LAMBDALABS_API_KEY"]


cli.add_command(instance_cmd.instance)
cli.add_command(ssh_cmd.ssh)
cli.add_command(fs_cmd.fs)


def cli_launch():
    cli(obj={})



import click
from . import cmd_utils as utl


@click.group()
@click.pass_context
@click.option("--json_output", is_flag=True, default=False)
def ssh(ctx, json_output):
    ctx.obj["JSON_OUTPUT"] = json_output


@click.command()
@click.pass_context
def list(ctx):
    utl.cli_message(ctx, "List SSH keys")
    ssh_keys = utl.do_get("ssh-keys", ctx)
    utl.cli_json_output(ssh_keys)


@click.command()
@click.pass_context
@click.option("--name", required=True, help="Name of the new ssh key to be created")
@click.option(
    "--public_key", default=None, help="Contents of the public key to be added"
)
def add(ctx, name, public_key):
    utl.cli_message(ctx, "Add SSH key")
    payload = {"name": name}
    if public_key:
        payload["public_key"] = public_key
    ssh_key = utl.do_post("ssh-keys", ctx, payload)
    utl.cli_json_output(ssh_key)


@click.command()
@click.pass_context
@click.option("--id", required=True, help="The id of ssh key to be deleted")
def delete(ctx, id):
    utl.cli_message(ctx, f"Deleting SSH key {id}")
    deleted_ssh_key = utl.do_delete(f"ssh-keys/{id}", ctx)
    utl.cli_json_output(deleted_ssh_key)


ssh.add_command(list)
ssh.add_command(add)
ssh.add_command(delete)



import click
from . import cmd_utils as utl


@click.group()
@click.pass_context
@click.option("--json_output", is_flag=True, default=False)
def fs(ctx, json_output):
    ctx.obj["JSON_OUTPUT"] = json_output


@click.command()
@click.pass_context
def list(ctx):
    utl.cli_message(ctx, "List File Systems")
    ssh_keys = utl.do_get("file-systems", ctx)
    utl.cli_json_output(ssh_keys)


fs.add_command(list)



from lamblbs.fs_cmd import fs
from click.testing import CliRunner
from unittest.mock import patch
import lamblbs.cmd_utils as utl
import json

API_KEY = "abc123"


@patch("lamblbs.cmd_utils.requests.get")
def test_list_fs_success(mock_get):
    mock_get.return_value.status_code = 200
    mock_get.return_value.json.return_value = {"data": []}

    runner = CliRunner()
    result = runner.invoke(fs, ["--json_output", "list"], obj={"API_KEY": API_KEY})

    response = json.loads(result.output)
    assert len(response["data"]) == 0


@patch("lamblbs.cmd_utils.requests.get")
def test_list_fs_fails(mock_get):
    mock_get.return_value.status_code = 401

    runner = CliRunner()
    result = runner.invoke(fs, ["--json_output", "list"], obj={"API_KEY": API_KEY})
    assert 1 == result.exit_code
    assert isinstance(result.exception, utl.APIException)
    assert result.exception.http_status == 401
